/* $Id: String.hh,v 1.1 2002/04/27 12:59:33 sdevaux Exp $
 *
 * NaJA
 * Tabs = 4 chars
 * copyright (C) 2002 Sebastien Devaux
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 *              Sebastien Devaux <sebastien.devaux@laposte.net>
 */ 

#ifndef __String_hh__
#define __String_hh__

#include "naja.hh"
#define MAX_INT 55555

PACKAGE_BEGIN(najaLang)

class String: public Object {
private:
	/** chaine vide par defaut */
	static char * null;
	/** longueur de la chaine */
	int len;
	/** Le contenu de la chaine proprement dit */
	char * value;
	/**
	 * Native c string length
	 * @param str string to monitor
	 * @return the lengh i.e. the count of char before the first \0 char
	 */ 
	int strlen(const char * str);
	/**
	 * Native string copy
	 * @param dest target string, enough memory must be allocated.
	 * @param src source string
	 * @param lmax max length to copy
	 */
	void strcpy(char * dest,const char * src,int lmax);
	/** 
	 * Native string compare
	 * @param s1 first string to compare
	 * @param s2 second string to compare
	 * @param lmax max length to compare (to full string comparaison, it must
	 *			include the shortest string's end char \0.
	 * @return <li>0 string are equals</li>
	 *		   <li>&gt;0 s1 > s2</li>
	 *		   <li>&lt;0 s1 < s2</li>
	 */
	int strcmp(const char * s1,const char * s2,int lmax);
	/** 
	 * Native string compare ignoring case
	 * @param s1 first string to compare
	 * @param s2 second string to compare
	 * @param lmax max length to compare (to full string comparaison, it must
	 *			include the shortest string's end char \0.
	 * @return <li>0 string are equals</li>
	 *		   <li>&gt;0 s1 > s2</li>
	 *		   <li>&lt;0 s1 < s2</li>
	 */
	int strcasecmp(const char * s1,const char * s2,int lmax);
	/**
	 * Create a new String from two Strings concatened
	 * @param str1 first string
	 * @param str2 second string
	 */
	String(const char * str1,const char * str2);
	/**
	 * Create a new String from another replacing one char with another
	 * @param str input String
	 * @param oldChar the old character
	 * @param newChar the new character
	 */
	String(NAJA<String> str,char oldChar,char newChar);
public:
	/**
     * Initializes a newly created String object so that it represents an empty
	 * character sequence.
	 */
	String();
	/**
	 * Free locally allocated memory
	 */
	~String();
	/** 
	 * Allocates a new String so that it represents the sequence of characters
	 * currently contained in the character array argument. The contents of the
     * character array are copied; subsequent modification of the character
	 * array does not affect the newly created string.
     * @param str the initial value of the string.
	 */
	String(const char * str);
	/**
	 * Allocates a new String that contains characters from a subarray of the
	 * character array argument. The offset argument is the index of the first
     * character of the subarray and the count argument specifies the length of
	 * the subarray. The contents of the subarray are copied; subsequent
     * modification of the character array does not affect the newly created string.
     * @param str array that is the source of characters.
     * @param offset the initial offset.
     * @param count the length.
	 */
	String(const char * str,int offset,int count);
	/**
	 * Returns the character at the specified index. An index ranges from 0 to
	 * length() - 1. The first character of the sequence is at index 0, the next
     * at index 1, and so on, as for array indexing.
	 * @param index the index of the character.
	 * @return the character at the specified index of this string. The first
	 *		 character is at index 0.
	 */
	virtual char charAt(int index);
	virtual int length();
	/**
	 * @see charAt
	 */
	virtual char operator[](int index);
	/**
	 * Copies characters from this string into the destination character array. 
     * The first character to be copied is at index srcBegin; the last character
	 * to be copied is at index srcEnd-1 (thus the total number of characters to
     * be copied is srcEnd-srcBegin). The characters are copied into the subarray
	 * of dst starting at index dstBegin and ending at index: 
     * <pre>dstBegin + (srcEnd-srcBegin) - 1</pre>
	 * @param srcBegin index of the first character in the string to copy.
     * @param srcEnd index after the last character in the string to copy.
     * @param dst the destination array.
     * @param dstBegin the start offset in the destination array.
	 */
	virtual void getChars(int srcBegin,int srcEnd,char * dst,int dstBegin);
	/**
	 * Compares this String to another String, ignoring case considerations. Two
     * strings are considered equal ignoring case if they are of the same
     * length, and corresponding characters in the two strings are equal ignoring case. 
     * Two characters c1 and c2 are considered the same, ignoring case if at least one
     * of the following is true:<ul> 
     *    <li>The two characters are the same (as compared by the == operator).</li>
     *    <li>Applying the method Character.toUpperCase(char) to each character produces
	 *		 the same result. 
     *	  <li>Applying the method Character.toLowerCase(char) to each character produces
	 *	     the same result.
	 * @param anotherString the String to compare this String against.
	 * @return true if the argument is not null and the Strings are equal, ignoring 
	 * 		case; false otherwise.
	 */
	virtual boolean equalsIgnoreCase(NAJA<String> anotherString);
	/**
	 * Compares two strings lexicographically. The character sequence
     * represented by this String object is compared lexicographically to the character
	 * sequence represented by the argument string. The result is a
     * negative integer if this String object lexicographically precedes the argument
	 * string. The result is a positive integer if this String object
     * lexicographically follows the argument string. The result is zero if the strings
	 * are equal; compareTo returns 0 exactly when the
     * equals(Object) method would return true.<br> 
     * This is the definition of lexicographic ordering. If two strings are different,
	 * then either they have different characters at some index that is a valid
     * index for both strings, or their lengths are different, or both. If they have
	 * different characters at one or more index positions, let k be the smallest
     * such index; then the string whose character at position k has the smaller value,
	 * as determined by using the < operator, lexicographically precedes
     * the other string. In this case, compareTo returns the difference of the two 
	 * character values at position k in the two string -- that is, the value:<ul> 
     *       <pre>this-&gt;charAt(k)-anotherString-&gt;charAt(k)</pre></ul>
     * If there is no index position at which they differ, then the shorter string
	 * lexicographically precedes the longer string. In this case, compareTo
     * returns the difference of the lengths of the strings -- that is, the value:<ul> 
     *       <pre>this.length()-anotherString.length()</pre></ul>
     * @param anotherString the String to be compared.
     * @return the value 0 if the argument string is equal to this string; a value less
	 *      than 0 if this string is lexicographically less than the string argument;
     *      and a value greater than 0 if this string is lexicographically greater than
	 *		 the string argument.
	 */
	virtual int compareTo(NAJA<String> anotherString);
	/**
	 * Compares two strings lexicographically, ignoring case considerations. This method
	 * returns an integer whose sign is that of
     * this.toUpperCase().toLowerCase().compareTo( str.toUpperCase().toLowerCase()).<br>
	 * Note that this method does not take locale into account, and will result in an
	 * unsatisfactory ordering for certain locales. The java.text package
     * provides collators to allow locale-sensitive ordering.
	 * @param anotherString the String to be compared.
	 * @return a negative integer, zero, or a positive integer as the the specified
	 * 		String is greater than, equal to, or less than this String, ignoring case
     *      considerations.
	 */
	virtual int compareToIgnoreCase(NAJA<String> anotherString);
	/**
	 * Tests if two string regions are equal. 
     * A substring of this String object is compared to a substring of the argument
	 * other. The result is true if these substrings represent identical
     * character sequences. The substring of this String object to be compared begins
	 * at index toffset and has length len. The substring of other to
     * be compared begins at index ooffset and has length len. The result is false if
	 * and only if at least one of the following is true:<ul> 
	 * 	<li>toffset is negative.</li> 
     *  <li>ooffset is negative.</li> 
     *  <li>toffset+len is greater than the length of this String object.</li> 
     *  <li>ooffset+len is greater than the length of the other argument.</li> 
     *  <li>There is some nonnegative integer k less than len such that:
	 *		this.charAt(toffset+k) != other.charAt(ooffset+k)</li>
	 * @param toffset the starting offset of the subregion in this string.
	 * @param other the string argument.
	 * @param ooffset the starting offset of the subregion in the string argument.
	 * @param l the number of characters to compare.
	 */
	virtual boolean regionMatches(int toffset,NAJA<String> other,
													int ooffset,int l);
	/**
	 * Tests if two string regions are equal. 
	 */
	virtual boolean regionMatch(boolean ignoreCase,int toffset,
							NAJA<String> other,int ooffset,int l);
	/**
	 * Tests if this string starts with the specified prefix beginning a 
	 * specified index.
	 */
	virtual boolean startsWith(NAJA<String> prefix,int offset);
	/**
	 * Tests if this string starts with the specified prefix.
	 */
	virtual boolean startsWith(NAJA<String> prefix);
	/**
	 * Tests if this string ends with the specified suffix.
	 */
	virtual boolean endsWith(NAJA<String> suffix);		
	/**
	 * Returns a hashcode for this string. The hashcode for a String object
	 * is computed as :
	 * <pre>s[0]*31^(n-1) + s[1]*31^(n-2) + ... + s[n-1]</pre>
	 */
	virtual int hashCode();
    /**
     * Returns the index within this string of the first occurrence of the
     * specified character, starting the search at the specified index.
     * <p>
     * If a character with value <code>ch</code> occurs in the character 
     * sequence represented by this <code>String</code> object at an index 
     * no smaller than <code>fromIndex</code>, then the index of the first
     * such occurrence is returned--that is, the smallest value <i>k</i> 
     * such that: 
     * <blockquote><pre>
     * (this.charAt(<i>k</i>) == ch) && (<i>k</i> >= fromIndex)
     * </pre></blockquote>
     * is true. If no such character occurs in this string at or after 
     * position <code>fromIndex</code>, then <code>-1</code> is returned.
     * <p>
     * There is no restriction on the value of <code>fromIndex</code>. If it 
     * is negative, it has the same effect as if it were zero: this entire 
     * string may be searched. If it is greater than the length of this 
     * string, it has the same effect as if it were equal to the length of 
     * this string: <code>-1</code> is returned.
     *
     * @param   ch          a character.
     * @param   fromIndex   the index to start the search from.
     * @return  the index of the first occurrence of the character in the
     *          character sequence represented by this object that is greater
     *          than or equal to <code>fromIndex</code>, or <code>-1</code>
     *          if the character does not occur.
     */
    virtual int indexOf(int ch, int fromIndex=0);
	/**
     * Returns the index within this string of the last occurrence of the
     * specified character, searching backward starting at the specified 
     * index. That is, the index returned is the largest value <i>k</i> 
     * such that:
     * <blockquote><pre>
     * this.charAt(k) == ch) && (k <= fromIndex)
     * </pre></blockquote>
     * is true.
     *
     * @param   ch          a character.
     * @param   fromIndex   the index to start the search from. There is no 
     *          restriction on the value of <code>fromIndex</code>. If it is 
     *          greater than or equal to the length of this string, it has 
     *          the same effect as if it were equal to one less than the 
     *          length of this string: this entire string may be searched. 
     *          If it is negative, it has the same effect as if it were -1: 
     *          -1 is returned.
     * @return  the index of the last occurrence of the character in the
     *          character sequence represented by this object that is less
     *          than or equal to <code>fromIndex</code>, or <code>-1</code>
     *          if the character does not occur before that point.
     */
    virtual int lastIndexOf(int ch, int fromIndex=MAX_INT);
	/**
	 * Returns the index within this string of the first occurrence of the
     * specified substring, starting at the specified index. The integer 
     * returned is the smallest value <i>k</i> such that:
     * <blockquote><pre>
     * this.startsWith(str, <i>k</i>) && (<i>k</i> >= fromIndex)
     * </pre></blockquote>
     * is <code>true</code>.
     * <p>
     * There is no restriction on the value of <code>fromIndex</code>. If 
     * it is negative, it has the same effect as if it were zero: this entire 
     * string may be searched. If it is greater than the length of this 
     * string, it has the same effect as if it were equal to the length of 
     * this string: <code>-1</code> is returned.
     *
     * @param   str         the substring to search for.
     * @param   fromIndex   the index to start the search from.
     * @return  If the string argument occurs as a substring within this
     *          object at a starting index no smaller than
     *          <code>fromIndex</code>, then the index of the first character
     *          of the first such substring is returned. If it does not occur
     *          as a substring starting at <code>fromIndex</code> or beyond,
     *          <code>-1</code> is returned.
     * @exception java.lang.NullPointerException if <code>str</code> is 
     *          <code>null</code>
     */
	virtual int indexOf(NAJA<String> str, int fromIndex=0);
    /**
     * Returns the index within this string of the last occurrence of
     * the specified substring.
     * The returned index indicates the start of the substring, and it
     * must be equal to or less than <code>fromIndex</code>. That is, 
     * the index returned is the largest value <i>k</i> such that:
     * <blockquote><pre>
     * this.startsWith(str, k) && (k <= fromIndex)
     * </pre></blockquote>
     * 
     * @param   str         the substring to search for.
     * @param   fromIndex   the index to start the search from. There is no 
     *          restriction on the value of fromIndex. If it is greater than 
     *          the length of this string, it has the same effect as if it 
     *          were equal to the length of this string: this entire string 
     *          may be searched. If it is negative, it has the same effect 
     *          as if it were -1: -1 is returned.
     * @return  If the string argument occurs one or more times as a substring
     *          within this object at a starting index no greater than
     *          <code>fromIndex</code>, then the index of the first character of
     *          the last such substring is returned. If it does not occur as a
     *          substring starting at <code>fromIndex</code> or earlier,
     *          <code>-1</code> is returned.
     * @exception java.lang.NullPointerException if <code>str</code> is 
     *          <code>null</code>.
     */
    virtual int lastIndexOf(NAJA<String> str, int fromIndex=MAX_INT);
    /**
     * Returns a new string that is a substring of this string. The 
     * substring begins at the specified <code>beginIndex</code> and 
     * extends to the character at index <code>endIndex - 1</code>. 
     * Thus the length of the substring is <code>endIndex-beginIndex</code>.
     * <p>
     * Examples:
     * <blockquote><pre>
     * "hamburger".substring(4, 8) returns "urge"
     * "smiles".substring(1, 5) returns "mile"
     * </pre></blockquote>
     *
     * @param      beginIndex   the beginning index, inclusive.
     * @param      endIndex     the ending index, exclusive.
     * @return     the specified substring.
     * @exception  IndexOutOfBoundsException  if the
     *             <code>beginIndex</code> is negative, or 
     *             <code>endIndex</code> is larger than the length of 
     *             this <code>String</code> object, or 
     *             <code>beginIndex</code> is larger than 
     *             <code>endIndex</code>.
     */
    virtual NAJA<String> substring(int beginIndex, int endIndex=MAX_INT);
	/**
     * Concatenates the specified string to the end of this string. 
     * <p>
     * If the length of the argument string is <code>0</code>, then this 
     * <code>String</code> object is returned. Otherwise, a new 
     * <code>String</code> object is created, representing a character 
     * sequence that is the concatenation of the character sequence 
     * represented by this <code>String</code> object and the character 
     * sequence represented by the argument string.<p>
     * Examples:
     * <blockquote><pre>
     * "cares".concat("s") returns "caress"
     * "to".concat("get").concat("her") returns "together"
     * </pre></blockquote>
     *
     * @param   str   the <code>String</code> that is concatenated to the end
     *                of this <code>String</code>.
     * @return  a string that represents the concatenation of this object's
     *          characters followed by the string argument's characters.
     * @exception java.lang.NullPointerException if <code>str</code> is 
     *          <code>null</code>.
     */
    virtual NAJA<String> concat(NAJA<String> str);
	/**
     * Returns a new string resulting from replacing all occurrences of 
     * <code>oldChar</code> in this string with <code>newChar</code>. 
     * <p>
     * If the character <code>oldChar</code> does not occur in the 
     * character sequence represented by this <code>String</code> object, 
     * then a reference to this <code>String</code> object is returned. 
     * Otherwise, a new <code>String</code> object is created that 
     * represents a character sequence identical to the character sequence 
     * represented by this <code>String</code> object, except that every 
     * occurrence of <code>oldChar</code> is replaced by an occurrence
     * of <code>newChar</code>. 
     * <p>
     * Examples:
     * <blockquote><pre>
     * "mesquite in your cellar".replace('e', 'o')
     *         returns "mosquito in your collar"
     * "the war of baronets".replace('r', 'y')
     *         returns "the way of bayonets"
     * "sparring with a purple porpoise".replace('p', 't')
     *         returns "starring with a turtle tortoise"
     * "JonL".replace('q', 'x') returns "JonL" (no change)
     * </pre></blockquote>
     *
     * @param   oldChar   the old character.
     * @param   newChar   the new character.
     * @return  a string derived from this string by replacing every
     *          occurrence of <code>oldChar</code> with <code>newChar</code>.
     */
    virtual NAJA<String> replace(char oldChar, char newChar);
	/**
     * Removes white space from both ends of this string. 
     * <p>
     * If this <code>String</code> object represents an empty character 
     * sequence, or the first and last characters of character sequence 
     * represented by this <code>String</code> object both have codes 
     * greater than <code>'&#92;u0020'</code> (the space character), then a 
     * reference to this <code>String</code> object is returned. 
     * <p>
     * Otherwise, if there is no character with a code greater than 
     * <code>'&#92;u0020'</code> in the string, then a new 
     * <code>String</code> object representing an empty string is created
     * and returned.
     * <p>
     * Otherwise, let <i>k</i> be the index of the first character in the 
     * string whose code is greater than <code>'&#92;u0020'</code>, and let 
     * <i>m</i> be the index of the last character in the string whose code 
     * is greater than <code>'&#92;u0020'</code>. A new <code>String</code> 
     * object is created, representing the substring of this string that 
     * begins with the character at index <i>k</i> and ends with the 
     * character at index <i>m</i>-that is, the result of 
     * <code>this.substring(<i>k</i>,&nbsp;<i>m</i>+1)</code>.
     * <p>
     * This method may be used to trim 
     * {@link Character#isSpace(char) whitespace} from the beginning and end 
     * of a string; in fact, it trims all ASCII control characters as well.
     *
     * @return  this string, with white space removed from the front and end.
     */
    virtual NAJA<String> trim();
	//---------------------------------------------------------
	// Object overrides
	/** 
	 * toString() override
	 * @return this string itself.
	 */
	virtual NAJA<String> toString();
	/**
	 * Compares this string to the specified object. The result is true if and only
     * if the argument is not null and is a String object that represents the
     * same sequence of characters as this object.
	 * @param anotherString the object to compare this String against.
	 * @return true if the String are equal; false otherwise.
	 */
	virtual boolean equals(NAJA<String> anotherString);  
	/**
	 * clone() override
	 * @return a copy of this string
	 */
	virtual NAJA<Object> clone();
	//---------------------------------------------------------
	// CCJLIB extensions
	/**
     * Acceder directement  la chaine native
	 * @return un pointeur const sur la representation interne de la
	 *		chaine
	 */
	virtual const char * c_str();
	//------------------------------------------------------
	// Surcharge operateurs
	friend NAJA<String> operator+(NAJA<String> str1,NAJA<String> str2);
	friend NAJA<String> operator+(NAJA<String> str1,const char * str2);
	friend NAJA<String> operator+(const char * str1,NAJA<String> str2);
};

extern "C++" {
NAJA<String> operator+(NAJA<String> str1,NAJA<String> str2);
NAJA<String> operator+(NAJA<String> str1,const char * str2);
NAJA<String> operator+(const char *,NAJA<String> str2);
}

PACKAGE_END

#endif // __String_hh__
